#include "../eSkel.c"   

// Parameters of the model
#define PROCS   8   // Number of processes, not used here but for information
#define STAGES  4   // Number of pipeline stages
        // !! when changing be careful with processes to stages affectation !!
#define INPUTNB 4   // Number of input data
#define INPUTSZ 1   // Size of the input data

// In this variant, there is a nested deal. Parameters of the deal:
// !! when changing any of these, be careful with processes to workers affectation !!
#define WORKERS 2   // Number of workers
#define DEALST  0   // ID of the stage which is a deal (0<=DEALST<STAGES)


// The stage procedure, taking as input a molecule, and returning a molecule
// In this example, it prints the input data and add 1000 to all the data
eSkel_molecule_t * MyStage (eSkel_molecule_t *thingy) {
  int i; 

  printf("MyStage -- len = %d, data = ", thingy->len[0]);  
  for (i=0;i<thingy->len[0];i++) {
    if (i<10) {
      printf("%.4d ", ((int *) thingy->data[0])[i]);
      ((int *) thingy->data[0])[i] += 1000;
    }
  }
  printf("\n");

  return thingy;
}


// The worker of the deal, taking as input a molecule, and returning a molecule
// In this example, it prints the input data and add 1000 to all the data
// (same job as a normal stage)
eSkel_molecule_t * DealWorker (eSkel_molecule_t *thingy) {
  int i; 

  printf("DealWorker -- len = %d, data = ", thingy->len[0]);  
  for (i=0;i<thingy->len[0];i++) {
    if (i<10) {
      printf("%.4d ", ((int *) thingy->data[0])[i]);
      ((int *) thingy->data[0])[i] += 1000;
    }
  }
  printf("\n");

  return thingy;
}


// This stage just performs a call to the Deal skeleton.
// It is persistently nested.
void StageDeal (void) {

  int workercol;
  int outmul;
  
  if ((myrank()<1))
    workercol = 0;
  else //if (myrank()==1)
    workercol = 1;
  //else
  // workercol = 2;

  Deal (WORKERS, IMPL, DealWorker, workercol, STRM,
	NULL, 0, 0, SPGLOBAL, MPI_INT,
	NULL, 0, &outmul, SPGLOBAL, MPI_INT, 0, mycomm());
}



// Main procedure, defining the Pipeline parameters and calling the Pipeline
int main (int argc, char *argv[])
{
  int i, j, p, next;
  spread_t spreads[STAGES+1];
  MPI_Datatype types[STAGES+1];
  Imode_t imodes[STAGES];
  int outmul;
  int mystagenum;

  int mymult;
  int *inputs, *results;

  eSkel_molecule_t *(*stages[STAGES])(eSkel_molecule_t *); 

  MPI_Init(&argc, &argv);
  SkelLibInit();

  // Welcome message
  if (myrank()==0) printf("*** Entering Example 1 ***\n");

  // Define the input data
  inputs = (int *) malloc (sizeof(int)*INPUTNB*INPUTSZ);
  for (i=0; i<INPUTNB*INPUTSZ; i++)
    inputs[i] = myrank()*100 + i;

  // Define the stage parameters
  for (i=0; i<STAGES+1; i++) {
    spreads[i] = SPGLOBAL;
    types[i]   = MPI_INT;
  }
  for (i=0; i<STAGES; i++) {
    imodes[i] = IMPL;
    stages[i] = (eSkel_molecule_t * (*)(eSkel_molecule_t *)) MyStage;
  }
  // Redefine the stage which is a Deal
  imodes[DEALST] = DEV;
  stages[DEALST] = (eSkel_molecule_t * (*)(eSkel_molecule_t *)) StageDeal;


  // Affect processes to stages
  // Be sure to affect enough processes to the Deal stage DEALST
  if (myrank()<1)
    mystagenum = 1;
  else if (myrank()<5)
    mystagenum = 0;
  else if ((myrank()==5)||(myrank()==7))
    mystagenum = 2;
  else
    mystagenum = 3;
  
  // Allocate the results buffer
  results = (int *) malloc (sizeof(int)*INPUTNB*INPUTSZ);

  // Call to the Pipeline function
  Pipeline (STAGES,  imodes, stages, mystagenum, BUF, spreads, types, 
  	    (void *) inputs, INPUTSZ, INPUTNB, (void *) results, INPUTSZ, &outmul,
  	    INPUTNB*INPUTSZ, mycomm());

  // Print out the results
  printf("Exiting Pipeline %d -- outmul: %d -- Results: ", myrank(), outmul); 
  for (i=0;i<INPUTSZ*INPUTNB;i++)
    printf("%d ",results[i]);
  printf("\n");

  // End MPI
  MPI_Finalize();
  return 0;
}
