// Some data marshalling operations which are not aware
// of anything to do with skeletons

// None of these are visible from the library API


void Local_Mult_Scatter (MPI_Comm c, int csz, int cid,     /* The group                                    */
                         int root,                         /* as an id within the group                    */
                         MPI_Datatype ty,                  /* type of object                               */
                         int tysz,                         /* size of the object sub-element in bytes      */
                         void **from, int frommul, 
                                      int *fromlen,        /* root inputs, multiplicity and lengths        */
                         char *to,  int *tomul,
                                    int *tolen)            /* local target buffers, tolen[0] is            */
                                                           /* multiplicity of preallocated buffer in units */
                                                           /* of size tysz, tomul returns actual           */
                                                           /* multiplicity, tolen[1] is length of          */
						           /* preallocated lengths buffer, overwritten     */
                                                           /* with actual lengths                          */
{
  int totalmul, totalsz;
  int tempsz;
  int *scatds, *scatls;
  int myinlen;   /* number of items expected */
  int myinspace; /* Copy of info provided in tolen[0] */
  int i, j;
  char *tempfrom, *fillptr;  /* for packing the from data struct from the final scatter */

  myinspace = tolen[0];

  /* Broadcast overall mult and compute own result mult */
  if (cid ==root) totalmul = frommul;
  MPI_Bcast(&totalmul, 1, MPI_INT, root, c);
  *tomul = totalmul/csz;
  if (cid < (totalmul%csz)) (*tomul)++;

  /* Scatter lens */
  if (tolen[1]<*tomul) {
    fprintf(stderr, "Local Mult Scatter but not enough space for lens\n");
    exit(0);
  }
  if (cid == root) {
    scatds = (int *) malloc(sizeof(int)*csz);
    scatls = (int *) malloc(sizeof(int)*csz);
    for (i=0; i<csz; i++) {
      if (i<(totalmul%csz)) {
        scatls[i] = totalmul/csz + 1;
      } else {
        scatls[i] = totalmul/csz;
      }
      if (i==0) {
        scatds[i] = 0;
      } else {
        scatds[i] = scatds[i-1]+scatls[i-1];
      }
    }
    MPI_Scatterv(fromlen, scatls, scatds, MPI_INT, (void *) tolen, *tomul, MPI_INT, root, c);
  } else {
     MPI_Scatterv(NULL, NULL, NULL, MPI_INT, (void *) tolen, *tomul, MPI_INT, root, c);
  }
    
  /* Scatter real data */
  myinlen = 0;
  for (i=0; i<*tomul; i++) {
    myinlen += tolen[i];
  }
  if (myinspace<myinlen) {
    fprintf(stderr, "Local Mult Scatter but not enough space for data\n");
    exit(0);
  }
  totalsz = 0;
  if (cid == root) {
    for (i=0; i<csz; i++) {
      tempsz = 0;
      for (j=scatds[i]; j<scatds[i]+scatls[i]; j++) {
        tempsz += fromlen[j];
      }
      scatls[i] = tempsz;
      totalsz += tempsz;
      if (i==0) {
        scatds[i] = 0;
      } else {
        scatds[i] = scatds[i-1]+scatls[i-1];
      }
    }
    tempfrom = (char *) malloc (totalsz*tysz);
    fillptr = tempfrom;
    for (i=0; i<totalmul; i++){
      memcpy ((void *) fillptr, (void *) from[i], fromlen[i]*tysz);
      fillptr += (fromlen[i]*tysz);
    }
    MPI_Scatterv(tempfrom, scatls, scatds, ty, to, myinlen, ty, root, c);
  } else {
    MPI_Scatterv(NULL, NULL, NULL, ty, to, myinlen, ty, root, c);
 }
}

void Local_Mult_Gather (MPI_Comm c, int csz, int cid,     /* The group                                    */
                        int root,                         /* as an id within the group                    */
                        MPI_Datatype ty,                  /* type of object                               */
                        int tysz,                         /* size of the object sub-element in bytes      */
                        void *from, int frommul, 
                                    int *fromlen,         /* local inputs, their multiplicity and lengths */
                        char ***to,  int *tomul,
			int **tolen)                      /* target buffer on the root, tomul returns     */
                                                          /* multiplicty                                  */
                                                          /* tolen returns lengths                        */
                                                          /* both to and tolen are dynamic alloc */
{
  int *allmul      = NULL;
  int *allindisp   = NULL;    /* at the root, displacements for final gather*/
  int *allinlen    = NULL;    /* at the root, lensfor final gather          */

  int *tempdisp    = NULL;
  char *temptobuff = NULL;    /* so that we can use a gather to get the to buffer stuff */
  char *cpptr;
  int i, j, k, temp;
  int totalmul     = 0;
  int totallen, mytotallen;

  /* root gathers mults     */
  if (cid == root) {
    allmul = (int *) malloc (sizeof(int)*csz);
  }
  temp = frommul;
  MPI_Gather(&temp, 1, MPI_INT, allmul, 1, MPI_INT, root, c);

  if (cid == root) {
    tempdisp = (int *) malloc (sizeof(int)*csz);
    tempdisp[0] = 0;
    for (i=0; i<csz; i++) {
        totalmul+=allmul[i];
        if (i>0) tempdisp[i] = tempdisp[i-1]+allmul[i-1];
    }
    *tolen = (int *)  malloc (totalmul*sizeof(int));
  }
  /* root gathervs tolens   */ 
  if (cid == root) {
    MPI_Gatherv(fromlen, frommul, MPI_INT, (void *) (*tolen), allmul, tempdisp, MPI_INT, root, c);
  } else {
    MPI_Gatherv(fromlen, frommul, MPI_INT, NULL, NULL, NULL, MPI_INT, root, c);
  }

  /* root gathervs the data */
  *to    = (char **) malloc (totalmul*sizeof(char *));
  mytotallen = 0;
  for (i=0; i<frommul; i++) {
    mytotallen+=fromlen[i];
  }
  if (cid==root) {
    totallen = 0;
    allindisp = (int *) malloc (sizeof(int)*csz);
    allinlen  = (int *) malloc (sizeof(int)*csz);
    allindisp[0] = 0;
    k = 0;
    for (i=0; i<csz; i++) {
      allinlen[i] = 0;
      for (j=0; j<allmul[i]; j++) {
        totallen    += (*tolen)[k];
        allinlen[i] += (*tolen)[k++];
      }
      if (i>0) allindisp[i] = allindisp[i-1]+allinlen[i-1];
    }

    temptobuff = (char *) malloc (totallen*tysz);
    MPI_Gatherv(from, mytotallen, ty, (void *) temptobuff, allinlen, allindisp, ty, root, c);
    *tomul = totalmul;
    /* now copy them into the permanent to buffer */
    cpptr = temptobuff;
    for (i=0; i<*tomul; i++) {
      (*to)[i] = (char *) malloc ((*tolen)[i]*tysz);
      memcpy ((void *) (*to)[i], (void *) cpptr, (*tolen)[i]*tysz);
      cpptr += ((*tolen)[i]*tysz);
    }
  } else {
    MPI_Gatherv(from, mytotallen, ty, NULL, NULL, NULL, ty, root, c);
  }
}

void Glob_Unit_Gather (MPI_Comm c, int csz, int cid,     /* The group                               */
                       int root,                         /* as an id within the group               */
                       MPI_Datatype ty,                  /* type of object                          */
                       int tysz,                         /* size of the object sub-element in bytes */
                       void *from, int fromlen,          /* local inputs                            */
                       char **to, int *tolen)            /* target buffer on the root, tolen is -1  */
                                                         /* if allocation required, else            */
                 					 /* length of preallocated buffer in units  */
                                                         /* of size tysz, returns actual length     */
{
  int *inlen  = NULL;    /* at the root, expected input lengths      */
  int *indisp = NULL;    /* at the root, corresponding displacements */
  int i, temp;
  int totallen;

  int size;
  MPI_Comm_size(c,&size);

  if (cid == root) {
    inlen  = (int *) malloc (sizeof(int)*csz);
    indisp = (int *) malloc (sizeof(int)*csz);
  }
  temp = fromlen;
  MPI_Gather(&temp, 1, MPI_INT, inlen, 1, MPI_INT, root, c);
  totallen=0;
  if (cid == root) {
      for (i=0; i<csz; i++) {
        totallen += inlen[i];
        if (i==0) indisp[i] = 0; else indisp[i] = indisp[i-1]+inlen[i-1];
      }
  }
  if (cid==root) {
    if (*tolen == -1) {
      *to = (char *) malloc (totallen*tysz);
    }
    else if (*tolen<totallen) {
      fprintf(stderr, "Global Unit Gather but not enough space\n");
      exit(0);
    }
    MPI_Gatherv(from, fromlen, ty, (void *)(*to), inlen, indisp, ty, root, c);
    *tolen = totallen;
  }
  else {
    MPI_Gatherv(from, fromlen, ty, NULL, NULL, NULL, ty, root, c);
  }
}

void Glob_Unit_Scatter (MPI_Comm c, int csz, int cid,     /* The group                               */
                        int root,                         /* as an id within the group               */
                        MPI_Datatype ty,                  /* type of object                          */
                        int tysz,                         /* size of the object sub-element in bytes */
                        void *from, int fromlen,          /* source buffer on the root               */
                        char **to, int *tolen)            /* dest local buffer, *tolen is -1         */
                                                          /* if allocation required, else            */
                   					  /* length of preallocated buffer in units  */
                                                          /* of size tysz, returns actual length     */
{
  int totallen, myshare;
  int *scatds, *scatls;
  int i, j;

  if (cid == root) totallen = fromlen;
  MPI_Bcast(&totallen, 1, MPI_INT, root, c);
  myshare = totallen/csz;
  if (cid < (totallen%csz)) myshare++;
  /* and now scatter this item */
  if (*tolen == -1) {
    *to = (char *) malloc (myshare*tysz);
  }
  else if (*tolen<myshare) {
    fprintf(stderr, "Global Unit Scatter but not enough space,  tolen %d, totallen %d, myshare %d\n", *tolen, totallen, myshare);
    exit(0);
  }
  *tolen = myshare;
  if  (cid == root) {
    /* set up the counts and displacements for the scatter */
    scatds = (int *) malloc(sizeof(int)*csz);
    scatls = (int *) malloc(sizeof(int)*csz);
    for (j=0; j<csz; j++) {
      if (j<(totallen%csz)) {
        scatls[j] = totallen/csz + 1;
      } else {
        scatls[j] = totallen/csz;
      }
      if (j==0) {
        scatds[j] = 0;
      } else {
        scatds[j] = scatds[j-1]+scatls[j-1];
      }
    }
  }
  if (cid == root) {
     MPI_Scatterv(from, scatls, scatds, ty, (void *) *to, myshare, ty, root, c);
  } else {
     MPI_Scatterv(NULL, NULL, NULL, ty, (void *) *to, myshare, ty, root, c);
  }
}


void group2group (MPI_Comm gc,          /* Global communicator          */
                  int gcsz,             /* its size                     */
                  int gcrank,           /* my rank in it                */
                  int role,             /* sender 0, receiver 1         */
                  int lrank,            /* my rank within sc or rc      */
                  MPI_Comm sc,          /* Senders communicator         */
       		                        /* NULL called from a receiver  */
                  int scsz,             /* its size                     */
                  int *srank,           /* sender ranks in gc           */
                  MPI_Comm rc,          /* Receivers communicator       */
       		                        /* NULL called from a sender    */
                  int rcsz,             /* its size                     */
                  int *rrank,           /* receiver ranks in gc         */
                  MPI_Datatype stype,   /* type of unit to send         */
                  int susz,             /* byte size of a sent unit     */    
                  char *sbuff,          /* sender buff                  */
                  int sblen,            /* send buffer length in units  */
                  char **rbuff,         /* receiver buff                */
                  int *rblen)           /* recv buffer length in units  */
                                        /* if pre-allocated, or -1      */
                                        /* if allocation required       */
                                        /* returns actual space filled  */

{
  char *tempsbuff;
  int tempsblen;
  char *temprbuff;
  int temprblen;
  MPI_Status stat;
  
  // first gather in the sending group
  if (role == SENDER) {
    tempsblen = -1;
    Glob_Unit_Gather (sc, scsz, lrank, 0, stype, susz, sbuff, sblen, &tempsbuff, &tempsblen);
  }

  // then a sending root to receiving root transfer
  // (could optimise this away if we arranged for the gather to go straight 
  //  here in the special case of overlapping send and receive groups
  if (role == SENDER && lrank == 0) {
    MPI_Send(&tempsblen, 1, MPI_INT, rrank[0], 0, gc);
    MPI_Send(tempsbuff, tempsblen, stype, rrank[0], 0, gc);
  } else if (role == RECEIVER && lrank == 0) {
    MPI_Recv(&temprblen, 1, MPI_INT, srank[0], MPI_ANY_TAG, gc, &stat);
    temprbuff = (char *) malloc (susz*temprblen);
    MPI_Recv(temprbuff, temprblen, stype, srank[0], MPI_ANY_TAG, gc, &stat);   
  }

  // finally scatter in the receiving group
  if (role == RECEIVER) {
     Glob_Unit_Scatter (rc, rcsz, lrank, 0, stype, susz, temprbuff, temprblen, rbuff, rblen);
  }

}

